"""

Example on deriving lidar moments from raw IQ data
--------------------------------------------------
"""

import highiq
import matplotlib.pyplot as plt

# Load an example ARM compliant-file
test_file = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)
print(test_file)
# Get the particle size distributions
out_ds = highiq.calc.get_psd(test_file)

# Plot the power spectra for a given time and height
fig, ax = plt.subplots(1, 1, figsize=(15, 5))
out_ds["power_spectral_density"].sel(range=350.0, method="nearest").plot(
    ax=ax,
    label="350 meters",
)
out_ds["power_spectral_density"].sel(range=950.0, method="nearest").plot(
    ax=ax,
    label="950 meters",
)
plt.legend()
plt.show()
plt.savefig("power_spectra_example.png")
test_file.close()



import numpy as np
import warnings
try:
    import cupy as cp
    cp.zeros(1)  # quick allocation test
    xp = cp
    print("Using CuPy (GPU)")
    from cupyx.scipy.signal import find_peaks
except Exception:
    import numpy as np
    xp = np
    print("Using NumPy (CPU)")
    from scipy.signal import find_peaks

import xarray as xr


def _gpu_calc_power(psd, dV, block_size=200, normed=True):
    shp = psd.shape
    power = xp.zeros((shp[0], shp[1]))
    if len(shp) == 3:
        gpu_array = xp.array(psd, dtype=xp.float32)
        if normed:
            gpu_array = gpu_array * dV
        gpu_array = xp.sum(gpu_array, axis=2)
        power = gpu_array.get() if hasattr(gpu_array, "get") else gpu_array
        
    else:
        gpu_array = xp.array(psd.values, dtype=xp.float32)
        if normed:
            gpu_array = gpu_array * dV
        gpu_array = xp.sum(gpu_array, axis=1)
        power = gpu_array.get() if hasattr(gpu_array, "get") else gpu_array
        
    return power


def _gpu_calc_velocity(psd, power, vel_bins, dV):
    shp = psd.shape
    gpu_array = xp.array(psd, dtype=xp.float32)
    power_array = xp.array(power, dtype=xp.float32)
    vel_bins_tiled = xp.tile(vel_bins, (shp[0], shp[1], 1))
    gpu_array = 1 / power_array * xp.sum(gpu_array * vel_bins_tiled, axis=2)
    velocity = gpu_array
    return velocity


def _gpu_calc_velocity_dumb(psd, vel_bins):
    dV = np.diff(vel_bins)[0]
    vel_min = vel_bins.min()
    gpu_array = xp.array(psd)
    gpu_array = xp.argmax(gpu_array, axis=2)
    gpu_array = vel_min + gpu_array.astype(xp.float32) * dV
    velocity = gpu_array
    return velocity


def _gpu_calc_spectral_width(psd, power, vel_bins, velocity, dV):
    shp = psd.shape
    times = shp[0]
    specwidth = xp.zeros((shp[0], shp[1]))
    gpu_array = xp.array(psd.values, dtype=xp.float32)
    power_array = xp.array(power, dtype=xp.float32)
    velocity_array = xp.transpose(xp.tile(velocity, (shp[2], 1, 1)), [1, 2, 0])
    vel_bins_tiled = xp.tile(vel_bins, (times, shp[1], 1))
    gpu_array = xp.sqrt(
        1
        / power_array
        * xp.sum((vel_bins_tiled - velocity_array) ** 2 * gpu_array, axis=2)
    )
    specwidth = gpu_array
    return specwidth


def _gpu_calc_skewness(psd, power, vel_bins, velocity, spec_width, dV):
    shp = psd.shape
    times = shp[0]
    gpu_array = xp.array(psd.values, dtype=xp.float32)
    power_array = xp.array(power, dtype=xp.float32)
    spec_width_array = xp.array(spec_width, dtype=xp.float32)
    power_array *= spec_width_array**3

    velocity_array = xp.transpose(xp.tile(velocity, (shp[2], 1, 1)), [1, 2, 0])
    vel_bins_tiled = xp.tile(vel_bins, (times, shp[1], 1))
    gpu_array = (
        1
        / power_array
        * xp.sum((vel_bins_tiled - velocity_array) ** 3 * gpu_array, axis=2)
    )
    skewness = gpu_array
    return skewness


def _gpu_calc_kurtosis(psd, power, vel_bins, velocity, spec_width, dV):
    shp = psd.shape
    kurtosis = np.zeros((shp[0], shp[1]))
    gpu_array = xp.array(psd.values, dtype=xp.float32)
    power_array = xp.array(power, dtype=xp.float32)
    spec_width_array = xp.array(spec_width, dtype=xp.float32)
    power_array *= spec_width_array**4
    velocity_array = xp.transpose(xp.tile(velocity, (shp[2], 1, 1)), [1, 2, 0])
    vel_bins_tiled = xp.tile(vel_bins, (shp[0], shp[1], 1))
    gpu_array = (
        1
        / power_array
        * xp.sum((vel_bins_tiled - velocity_array) ** 4 * gpu_array, axis=2)
    )
    kurtosis = gpu_array
    return kurtosis


def get_lidar_moments(
    spectra, intensity_thresh=0, block_size_ratio=1.0, which_moments=None
):
    """
    This function will retrieve the lidar moments of the Doppler spectra.

    Parameters
    ----------
    spectra: ACT Dataset
        The dataset containing the processed Doppler spectral density functions.
    intensity_thresh: float
        The minimum signal to noise ratio to use as an initial mask of noise.
    block_size_ratio: float
        This value is used to determine how much data the GPU will process in one loop. If your
        GPU has more memory, you may be able to optimize processing by raising this number. In
        addition, if you encounter out of memory errors, try lowering this number, ensuring that
        it is a positive floating point number.
    which_moments: list or None
        This tells HighIQ which moments should be processed. If this list is None, then the
        signal to noise ratio, doppler velocity, spectral width, skewness,
        and kurtosis will be calculated.

    Returns
    -------
    spectra: ACT Dataset
        The database with the Doppler lidar moments.
    """
    if "power_spectral_density" not in spectra.variables.keys():
        raise ValueError(
            "You must calculate the power spectra before calculating moments!"
        )
    if which_moments is None:
        which_moments = [
            "intensity",
            "radial_velocity",
            "spectral_width",
            "skewness",
            "kurtosis",
        ]
    else:
        which_moments = [x.lower() for x in which_moments]

    if not block_size_ratio > 0:
        raise ValueError("block_size_ratio must be a positive floating point number!")

    dV = np.diff(spectra["vel_bins"])[0]
    linear_psd = spectra["power_spectral_density"] - 1
    linear_psd_0filled = linear_psd.fillna(0).values
    power = _gpu_calc_power(linear_psd_0filled, dV)
    velocity = _gpu_calc_velocity(
        linear_psd_0filled, power, spectra["vel_bins"].values, dV
    )

    if "intensity" in which_moments:
        power_with_noise = dV * spectra["power_spectral_density"].sum(dim="vel_bins")
        spectra["intensity"] = power_with_noise / (dV * len(spectra["vel_bins"]))
        spectra["intensity"].attrs["long_name"] = "Signal to Noise Ratio + 1"
        spectra["intensity"].attrs["units"] = ""
        spectra.attrs["intensity_mask"] = "%f" % intensity_thresh

    if "radial_velocity" in which_moments:
        velocity_dumb = _gpu_calc_velocity_dumb(
            linear_psd_0filled, spectra["vel_bins"].values
        )
        velocity_dumb = velocity_dumb.get() if hasattr(velocity_dumb, "get") else velocity_dumb
        spectra["radial_velocity_max_peak"] = xr.DataArray(
            velocity_dumb, dims=("time", "range")
        )
        spectra["radial_velocity_max_peak"].attrs[
            "long_name"
        ] = "Doppler velocity derived using location of highest peak in spectra."
        spectra["radial_velocity_max_peak"].attrs["units"] = "m s-1"
        velocity = velocity.get() if hasattr(velocity, "get") else velocity
        spectra["radial_velocity"] = xr.DataArray(velocity, dims=("time", "range"))
        spectra["radial_velocity"].attrs[
            "long_name"
        ] = "Doppler velocity using first moment"
        spectra["radial_velocity"].attrs["units"] = "m s-1"
        spectra["radial_velocity_max_peak"] = spectra["radial_velocity_max_peak"].where(
            spectra.intensity > intensity_thresh
        )
        spectra["radial_velocity"] = spectra["radial_velocity"].where(
            spectra.intensity > intensity_thresh
        )

    if (
        "spectral_width" in which_moments
        or "kurtosis" in which_moments
        or "skewness" in which_moments
    ):
        spectral_width = _gpu_calc_spectral_width(
            linear_psd, power, spectra["vel_bins"].values, velocity, dV
        )

    if "spectral_width" in which_moments:
        spectral_width = spectral_width.get() if hasattr(spectral_width, "get") else spectral_width
        spectra["spectral_width"] = xr.DataArray(spectral_width, dims=("time", "range"))
        spectra["spectral_width"].attrs["long_name"] = "Spectral width"
        spectra["spectral_width"].attrs["units"] = "m s-1"
        if "intensity" in which_moments:
            spectra["spectral_width"] = spectra["spectral_width"].where(
                spectra.intensity > intensity_thresh
            )

    if "skewness" in which_moments:
        skewness = _gpu_calc_skewness(
            linear_psd, power, spectra["vel_bins"].values, velocity, spectral_width, dV
        )
        skewness = skewness.get() if hasattr(skewness, "get") else skewness
        spectra["skewness"] = xr.DataArray(
            skewness, dims=("time", "range"))
        if "intensity" in which_moments:
            spectra["skewness"] = spectra["skewness"].where(
                spectra.intensity > intensity_thresh
            )
        spectra["skewness"].attrs["long_name"] = "Skewness"
        spectra["skewness"].attrs["units"] = "m^3 s^-3"

    if "kurtosis" in which_moments:
        kurtosis = _gpu_calc_kurtosis(
            linear_psd, power, spectra["vel_bins"].values, velocity, spectral_width, dV
        )
        kurtosis = kurtosis.get() if hasattr(kurtosis, "get") else kurtosis
        spectra["kurtosis"] = xr.DataArray(
            kurtosis, dims=("time", "range"))
        if "intensity" in which_moments:
            spectra["kurtosis"] = spectra["kurtosis"].where(
                spectra.intensity > intensity_thresh
            )
        spectra["kurtosis"].attrs["long_name"] = "Kurtosis"
        spectra["kurtosis"].attrs["units"] = "m^4 s^-4"

    spectra["vel_bins"].attrs["long_name"] = "Doppler velocity"
    spectra["vel_bins"].attrs["units"] = "m s-1"

    return spectra



"""

Example on plotting moments from raw IQ data
---------------------------------------------
"""

import highiq
import matplotlib.pyplot as plt

# Load an example ARM compliant-file
test_file = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)

# Get the particle size distributions
my_ds = highiq.calc.get_psd(test_file)
my_ds = highiq.calc.get_lidar_moments(my_ds)
print(my_ds)
# Filter dataset based on SNR
my_ds["radial_velocity_max_peak"] = my_ds["radial_velocity_max_peak"].where(
    my_ds.intensity > 1.008
)

# Plot the power spectra for a given time and height
fig, ax = plt.subplots(1, 2, figsize=(15, 5))
my_ds["intensity"].plot(ax=ax[0], y="range")
my_ds["radial_velocity_max_peak"].plot(ax=ax[1], y="range")
plt.show()
plt.savefig("moments_example.png")
test_file.close()



import numpy as np
import warnings

try:
    import cupy as cp

    cp.zeros(1)  # quick allocation test
    xp = cp
    print("Using CuPy (GPU)")
    from cupyx.scipy.signal import find_peaks
except Exception:
    import numpy as np

    xp = np
    print("Using NumPy (CPU)")
    from scipy.signal import find_peaks
import xarray as xr

from pint import UnitRegistry


def _fast_expand(complex_array, factor, num_per_block=200):
    shp = complex_array.shape
    expanded_array = np.zeros((shp[0], shp[1], shp[2] * factor))
    weights = np.tile(np.arange(0, factor) / factor, (shp[0], shp[1], 1))
    for i in range(shp[2]):
        gpu_array = xp.zeros(complex_array[:, :, i], dtype=xp.float32)
        if i < shp[2] - 1:
            gpu_array2 = xp.zeros(complex_array[:, :, i + 1], dtype=xp.float32)
            diff_array = gpu_array2 - gpu_array
        else:
            diff_array = xp.zeros((shp[0], shp[1]))

        rep_array = xp.transpose(xp.tile(gpu_array, (factor, 1, 1)), [1, 2, 0])
        diff_array = xp.transpose(xp.tile(diff_array, (factor, 1, 1)), [1, 2, 0])
        temp_array = diff_array * weights + rep_array
        expanded_array[:, :, factor * i : factor * (i + 1)] = np.array(temp_array)
    return expanded_array


def get_psd(
    spectra,
    gate_resolution=60.0,
    wavelength=None,
    fs=None,
    nfft=1024,
    time_window=10,
    acf_name="acf",
    acf_bkg_name="acf_bkg",
    nsamples=4000,
    hamming_window=True,
):
    """
    This function will get the power spectral density from the autocorrelation function.

    Parameters
    ----------
    spectra: ACT Dataset
        The dataset containing the autocorrelation function.
    gate_resolution: float
        The gate resolution to derive the spectra for.
    wavelength: float or None
        The wavelength (in m) of the radar. If None, HighIQ will attempt to load the
        wavelength from the wavelength attribute of the spectra dataset.
    fs: float or None
        The pulse repetition frequency of the radar in Hz. If None,
        HighIQ will try and extract this information from the
        ACT dataset. This will require that the dataset
        have an attribute called sample_rate that contains a string
        with the magnitude and units of the sample rate.
    nfft: int
        The number of points to include in the FFT.
    time_window: int
        The number of time points to include in the rolling window.
    acf_name: str
        The name of the autocorrelation function field.
    acf_bkg_name: str
        The name of the autocorrelation function of the background.
    nsamples: int
        Number of samples in ACF
    hamming_window: bool
        Whether to apply a Hamming window to the data before performing the FFT.

    Returns
    -------
    spectra: ACT Dataset
        The dataset containing the power spectral density functions.
    """

    Q_ = UnitRegistry().Quantity
    if fs is None:
        if "sample_rate" not in spectra.attrs:
            raise KeyError(
                "If sample frequency is not specified, then ACT Dataset must contain a sample_rate attribute!"
            )

        fs_pint = Q_(spectra.attrs["sample_rate"])
        fs_pint = fs_pint.to("Hz")
        fs = fs_pint.magnitude

    if wavelength is None:
        if "wavelength" not in spectra.attrs:
            raise KeyError(
                "If wavelength is not specified, then the dataset must contain a wavelength attribute!"
            )
        fs_pint = Q_(spectra.attrs["wavelength"])
        fs_pint = fs_pint.to("m")
        wavelength = fs_pint.magnitude

    spectra_out = spectra.resample(time="%ds" % time_window).mean()
    acf = spectra_out[acf_name]
    num_times = acf.shape[0]
    num_gates = int(gate_resolution / 3)
    complex_coeff_in = acf.isel(complex=0).values + acf.isel(complex=1).values * 1j
    complex_coeff_in = xp.array(complex_coeff_in, dtype=xp.complex128)
    complex_coeff = xp.zeros(
        (
            num_times,
            int(complex_coeff_in.shape[1] / num_gates),
            complex_coeff_in.shape[2],
        ),
        dtype=xp.complex128,
    )
    for i in range(complex_coeff.shape[1]):
        complex_coeff[:, i, :] = xp.mean(
            complex_coeff_in[:, (num_gates * i) : (num_gates * (i + 1)), :], axis=1
        )
    complex_coeff_in = None
    complex_coeff = complex_coeff / xp.tile(
        complex_coeff[:, :, 0], (complex_coeff.shape[2], 1, 1)
    ).transpose(1, 2, 0)

    if hamming_window:
        # Only use right half of the Hamming window
        hamming_win = xp.hamming(complex_coeff.shape[2] * 2)[complex_coeff.shape[2] :]
        hamming_win = xp.tile(
            hamming_win, (complex_coeff.shape[0], complex_coeff.shape[1], 1)
        )
        complex_coeff = complex_coeff * hamming_win

    freq = np.fft.fftshift(np.fft.fftfreq(nfft) * fs)
    spectra_out.attrs["nyquist_velocity"] = "%f m s-1" % (wavelength / (4 * 1 / fs))
    spectra_out["freq_bins"] = xr.DataArray(freq, dims=["freq"])
    spectra_out["freq_bins"].attrs[
        "long_name"
    ] = "Doppler spectra bins in frequency units"
    spectra_out["freq_bins"].attrs["units"] = "Hz"
    vel_bins = spectra_out["freq_bins"] * (wavelength / 2)
    inds_sorted = np.argsort(vel_bins.values)
    vel_bins = vel_bins[inds_sorted]
    attrs_dict = {"long_name": "Doppler velocity", "units": "m s-1"}
    spectra_out["vel_bins"] = xr.DataArray(
        vel_bins, dims=("vel_bins"), attrs=attrs_dict
    )
    spectra_out["freq_bins"] = spectra_out["freq_bins"][inds_sorted]

    complex_coeff_bkg_in = (
        spectra_out[acf_bkg_name].isel(complex=0).values
        + spectra_out[acf_bkg_name].isel(complex=1).values * 1j
    )
    complex_coeff_bkg_in = xp.array(complex_coeff_bkg_in, dtype=xp.complex128)
    complex_coeff_bkg = xp.zeros(
        (
            num_times,
            int(complex_coeff_bkg_in.shape[1] / num_gates),
            complex_coeff_bkg_in.shape[2],
        ),
        dtype=xp.complex128,
    )
    for i in range(complex_coeff.shape[1]):
        complex_coeff_bkg[:, i, :] = xp.mean(
            complex_coeff_bkg_in[:, (num_gates * i) : (num_gates * (i + 1)), :], axis=1
        )
    complex_coeff_bkg = complex_coeff_bkg / xp.tile(
        complex_coeff_bkg[:, :, 0], (complex_coeff_bkg.shape[2], 1, 1)
    ).transpose(1, 2, 0)
    if hamming_window:
        hamming_win = xp.hamming(complex_coeff.shape[2] * 2)[complex_coeff.shape[2] :]
        hamming_win = xp.tile(
            hamming_win, (complex_coeff_bkg.shape[0], complex_coeff_bkg.shape[1], 1)
        )
        complex_coeff_bkg = complex_coeff_bkg * hamming_win

    num_lags = complex_coeff_bkg_in.shape[2]
    if nfft < num_lags:
        raise RuntimeError("Number of points in FFT < number of lags in sample!")
    frames = xp.asarray(complex_coeff)
    arr = xp.fft.fft(frames, n=nfft)
    power = xp.abs(xp.fft.fftshift(arr, axes=2))
    power = power.get() if hasattr(power, "get") else power

    attrs_dict = {"long_name": "Range", "units": "m"}
    spectra_out["range"] = xr.DataArray(
        gate_resolution * np.arange(power.shape[1]),
        dims=("range"),
        attrs=attrs_dict,
    )
    frames = xp.asarray(complex_coeff_bkg)
    arr = xp.fft.fft(frames, n=nfft)
    power_bkg = xp.abs(xp.fft.fftshift(arr, axes=2))
    power_bkg = power_bkg.get() if hasattr(power_bkg, "get") else power_bkg
    # Subtract background noise
    spectra_out["power_spectral_density"] = (
        ["time", "range", "vel_bins"],
        power / power_bkg,
    )

    # Ground noise floor to 1
    spectra_out["power_spectral_density"].attrs["long_name"] = "Power spectral density"
    spectra_out["power_spectral_density"].attrs["units"] = ""
    spectra_out["range"].attrs["long_name"] = "Range"
    spectra_out["range"].attrs["units"] = "m"
    spectra_out["vel_bins"].attrs["long_name"] = "Doppler velocity"
    spectra_out["vel_bins"].attrs["units"] = "m s-1"
    power = None
    power_bkg = None
    spectra_out = spectra_out.drop(["acf", "acf_bkg"])
    return spectra_out


def calc_num_peaks(my_spectra, **kwargs):
    """

    This function will calculate the number of peaks in the spectra.

    Parameters
    ----------
    my_spectra: ACT Dataset
        The dataset to calculate the number of peaks for.
    kwargs:
        Additional keyword arguments are passed into :func:`scipy.signal.find_peaks`.
        The default minimum height and width of the peak are set to 3 and 8 points
        respectively.

    Returns
    -------
    my_spectra: ACT Dataset
        The dataset with an 'npeaks' variable included that shows the number of peaks.
    """
    spectra = my_spectra["power_spectral_density"]
    my_array = xp.array(spectra.fillna(0).values)
    shp = my_array.shape
    num_peaks = xp.zeros((shp[0], shp[1]))
    peak_loc = xp.nan * xp.ones((shp[0], shp[1], 5))

    vel_bins = my_spectra["vel_bins"].values

    if "prominence" not in kwargs.keys():
        prominence = 0.01
    else:
        prominence = kwargs.pop("prominence")

    if "width" not in kwargs.keys():
        width = 8
    else:
        width = kwargs.pop("width")

    if "height" not in kwargs.keys():
        height = 1.5
    else:
        height = kwargs.pop("height")

    for i in range(shp[0]):
        for j in range(shp[1]):
            peaks = find_peaks(
                my_array[i, j],
                prominence=prominence,
                width=width,
                height=height,
                **kwargs,
            )[0]
            peaks = peaks.get() if hasattr(peaks, "get") else peaks
            num_peaks[i, j] = len(peaks)
            for k in range(len(peaks)):
                if k > 4:
                    continue
                peak_loc[i, j, k] = vel_bins[peaks[k]]
    num_peaks = num_peaks.get() if hasattr(num_peaks, "get") else num_peaks
    peak_loc = peak_loc.get() if hasattr(peak_loc, "get") else peak_loc
    my_spectra["npeaks"] = xr.DataArray(num_peaks, dims=("time", "range"))
    my_spectra["npeaks"].attrs["long_name"] = "Number of peaks in Doppler spectra"
    my_spectra["npeaks"].attrs["units"] = "1"
    my_spectra["peak_velocities"] = xr.DataArray(
        peak_loc, dims=("time", "range", "peak_no")
    )
    my_spectra["peak_velocities"].attrs["long_name"] = "Dopper velocity peaks"
    my_spectra["peak_velocities"].attrs["units"] = "m s-1"
    return my_spectra



import os
import struct
import pandas as pd
import xarray as xr

from datetime import datetime

import numpy as np

global_attrs = {
    "number_of_lags": {"cor.M1": 7, "hou.M1": 7, "oli.M1": 7, "sgp.C1": 20},
    "number_of_samples_per_lag": {
        "cor.M1": 3200,
        "hou.M1": 3200,
        "oli.M1": 1920,
        "sgp.C1": 4000,
    },
}


def load_arm_netcdf(arm_file, **kwargs):
    """

    This loads netCDF data that are in the Atmospheric Radiation Measurement standard netCDF format.

    Parameters
    ----------
    arm_file: str
        The path to the dataset to load.

    Additional keyword arguments are passed into :func`xr.open_dataset`

    Returns
    -------
    ds: Xarray Dataset
        Returns the xarray dataset that contains the autocorrelation functions.
    """

    ds = xr.open_dataset(arm_file, **kwargs)
    if "time" not in ds["acf"].dims:
        ds["acf"] = ds["acf"].expand_dims("time")
    if "time" not in ds["acf_bkg"].dims:
        ds["acf_bkg"] = ds["acf_bkg"].expand_dims(dim={"time": len(ds.time)}, axis=0)
    return ds


def read_00_data(
    file_name, home_point, site="sgp.C1", sample_start=None, sample_end=None, **kwargs
):
    """
    This reads a raw StreamLine Doppler Lidar .00 level file.

    Parameters
    ----------
    file_name: str
        Name of .raw file to read
    home_point: str
        Name of home point file path.
    sample_start: int
        Start of time period to load (beam index)
    sample_end: int
        End of time periods to load (beam index)
    kwargs

    Returns
    -------
    ds: ACT Dataset
        Returns the ACT dataset (xarray dataset) that contains the autocorrelation functions.
    """
    with open(file_name, mode="rb") as f:
        home_point_columns = [
            "start_time (yyyymmddhh)",
            "end_time (yyyymmddhh)",
            "Target_latitude",
            "Target_longitude",
            "target_altitude",
            "lidar_latitude",
            "lidar_longitude",
            "lidar_altitude",
            "lidar_home_point",
            "Descriptive comment",
        ]
        num_rows_to_skip = 0
        with open(home_point) as hf:
            for line in hf:
                if line[0] == "#":
                    num_rows_to_skip += 1

        hfile = pd.read_csv(
            home_point, skiprows=num_rows_to_skip, names=home_point_columns
        )
        background_vals = (
            global_attrs["number_of_lags"][site]
            * global_attrs["number_of_samples_per_lag"][site]
            * 2
        )
        nlags = global_attrs["number_of_lags"][site]
        num_samples = global_attrs["number_of_samples_per_lag"][site]
        background_bytes = 8 * background_vals
        beam_bytes = 24 + background_bytes
        dot_split = file_name.split(".")
        if len(dot_split) < 3:
            dot_split = dot_split[0].split("_")
        for x in dot_split:
            if len(x) == 8 and x.isnumeric():
                date_str = x
            if len(x) == 6 and x.isnumeric():
                time_str = x
            under_split = x.split("_")
            for x in under_split:
                print(x)
                if len(x) == 8 and x.isnumeric():
                    date_str = x
                if len(x) == 2 and x.isnumeric():
                    time_str = x + "0000"

        base_time = datetime.strptime("%s.%s" % (date_str, time_str), "%Y%m%d.%H%M%S")
        midnight = datetime(
            year=base_time.year,
            month=base_time.month,
            day=base_time.day,
            hour=0,
            minute=0,
            second=0,
        ).timestamp()
        base_time_int = int(base_time.strftime("%Y%m%d%H%M%S"))
        for i in range(1, len(hfile)):
            num = int(hfile["start_time (yyyymmddhh)"][i])
            end_num = int(hfile["end_time (yyyymmddhh)"][i])
            if end_num == -1:
                end_num = np.inf
            if base_time_int >= num and base_time_int < end_num:
                target_latitude = float(hfile["Target_latitude"][i])
                target_longitude = float(hfile["Target_longitude"][i])
                target_altitude = float(hfile["target_altitude"][i])
                lidar_latitude = float(hfile["lidar_latitude"][i])
                lidar_longitude = float(hfile["lidar_longitude"][i])
                lidar_home_point = float(hfile["lidar_home_point"][i])

        if site[:3] == "oli":
            read_background = False
            total_sample_bytes = os.path.getsize(file_name)
        else:
            read_background = True
            total_sample_bytes = os.path.getsize(file_name) - background_bytes

        if sample_start is None and sample_end is None:
            nsamples = int(total_sample_bytes / beam_bytes)
        else:
            maxnsamples = int(total_sample_bytes / beam_bytes)
            if sample_end > maxnsamples:
                sample_end = maxnsamples
            nsamples = sample_end - sample_start

        if read_background:
            stringFmt = "<"
            for i in range(background_vals):
                stringFmt += "d"
            my_data = f.read(background_bytes)
            acf_bkg = np.array(struct.unpack(stringFmt, my_data))
            acf_bkg = np.reshape(acf_bkg, (nlags, num_samples, 2))
            acf_bkg = np.transpose(acf_bkg, [1, 0, 2])

        azimuth = np.zeros(nsamples)
        elevation = np.zeros(nsamples)
        timestamp = np.zeros(nsamples)
        acf = np.zeros((nsamples, num_samples, nlags, 2))

        stringFmt = "<"
        for i in range(background_vals):
            stringFmt += "d"
        previous_time = 0
        spill_over = 0.0
        if sample_start is not None:
            num_bytes_to_skip = beam_bytes * sample_start
            f.seek(num_bytes_to_skip, 1)

        for i in range(nsamples):
            data = f.read(8)
            if len(data) < 8:
                break
            azimuth[i] = float(struct.unpack("<d", data)[0]) + lidar_home_point
            data = f.read(8)
            elevation[i] = float(struct.unpack("<d", data)[0])
            data = f.read(8)
            timestamp[i] = float(struct.unpack("<d", data)[0])

            seconds_decimal = timestamp[i] * 3600
            seconds = int(seconds_decimal)
            seconds_fraction = seconds_decimal - float(seconds)
            seconds_since_1970 = seconds + midnight + seconds_fraction
            timestamp[i] = seconds_since_1970
            if previous_time == 0:
                previous_time = seconds_decimal
            if previous_time - seconds_decimal > 86400:
                spill_over += 86400
            previous_time = seconds_decimal
            seconds_since_1970 += spill_over

            data = f.read(background_vals * 8)
            acf_temp = np.array(struct.unpack(stringFmt, data))
            acf_temp = np.reshape(acf_temp, (nlags, num_samples, 2))
            acf_temp = np.transpose(acf_temp, [1, 0, 2])
            acf[i] = acf_temp

        azimuth = xr.DataArray(azimuth.astype(np.float32), dims=("time"))
        azimuth.attrs["long_name"] = "Azimuth angle"
        azimuth.attrs["units"] = "Degree"
        elevation = xr.DataArray(elevation.astype(np.float32), dims=("time"))
        elevation.attrs["units"] = "Degree"
        elevation.attrs["long_name"] = "Elevation angle"
        timestamp = np.array([datetime.fromtimestamp(x) for x in timestamp])
        timestamp = xr.DataArray(timestamp, dims=("time"))

        acf_bkg = xr.DataArray(
            acf_bkg.astype(np.float32), dims=("nsamples", "nlags", "complex")
        )
        acf_bkg.attrs["long_name"] = "Autocorrelation function of background"
        acf = xr.DataArray(
            acf.astype(np.float32), dims=("time", "nsamples", "nlags", "complex")
        )
        acf.attrs["long_name"] = "Autocorrelation function"
        lidar_ds = xr.Dataset(
            {
                "acf": acf,
                "acf_bkg": acf_bkg,
                "aziumth": azimuth,
                "time": timestamp,
                "elevation": elevation,
            }
        )
        lidar_ds = lidar_ds.sortby("time")
        lidar_ds.attrs["dlon"] = lidar_longitude
        lidar_ds.attrs["dlat"] = lidar_latitude
        lidar_ds.attrs["target_latitude"] = target_latitude
        lidar_ds.attrs["target_longitude"] = target_longitude
        lidar_ds.attrs["target_altitude"] = target_altitude
        lidar_ds.attrs["home_point_azimuth"] = lidar_home_point
        lidar_ds.attrs["sample_rate"] = "50 MHz"
        lidar_ds.attrs["wavelength"] = "1548 nm"
    return lidar_ds



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import sphinx_theme

# -- Project information -----------------------------------------------------

project = "HighIQ"
copyright = "2019, Bobby Jackson"
author = "Bobby Jackson"

# The full version, including alpha/beta/rc tags
release = "1.2"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_gallery.gen_gallery",
]

# -- General configuration ---------------------------------------------------

sphinx_gallery_conf = {
    "examples_dirs": "../../examples",
    "gallery_dirs": "source/auto_examples",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]  # your stylesheet(s) inside _static/css/
html_theme_path = [sphinx_theme.get_html_theme_path()]

# Numpy autodoc attributes
numpydoc_show_class_members = True
autosummary_generate = True



import highiq
import numpy as np


def test_io():
    my_ds = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)
    assert "acf" in my_ds.variables.keys()
    assert "acf_bkg" in my_ds.variables.keys()
    my_ds.close()


def test_spectra():
    my_ds = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)
    my_spectra = highiq.calc.get_psd(my_ds)
    assert "power_spectral_density" in my_spectra.variables.keys()
    psd = my_spectra["power_spectral_density"].sel(range=400, method="nearest")
    vel_bins = my_spectra["vel_bins"]
    dV = vel_bins[1] - vel_bins[0]
    np.testing.assert_almost_equal(
        psd.values.sum() * dV.values, 39.644146692612104, decimal=1
    )
    my_ds.close()
    my_spectra.close()


def test_moments():
    my_ds = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)
    my_spectra = highiq.calc.get_psd(my_ds)
    my_moments = highiq.calc.get_lidar_moments(my_spectra)
    intensity = my_moments["intensity"].values
    velocity = my_moments["radial_velocity"].values
    assert np.nanmin(intensity) > -1.0
    assert np.nanmin(velocity) < -2.0
    my_ds.close()
    my_spectra.close()


def test_peaks():
    my_ds = highiq.io.load_arm_netcdf(highiq.testing.TEST_FILE)
    my_spectra = highiq.calc.get_psd(my_ds, gate_resolution=200.0)
    my_peaks = highiq.calc.calc_num_peaks(
        my_spectra, height=1.5, width=8, prominence=0.5
    )
    my_peaks = highiq.calc.get_lidar_moments(my_peaks)
    my_peaks["npeaks"] = my_peaks["npeaks"].where(my_peaks.intensity > 0.5)
    num_peaks = my_peaks["npeaks"].values
    assert np.nanmax(num_peaks) == 1
    my_ds.close()
    my_spectra.close()
    my_peaks.close()
